/*
	Copyright (C) 2012 AJA Video Systems, Inc.  Proprietary and Confidential information.  All rights reserved.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
	TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
	THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
	TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef __AJA_CONTROL_H__
#define __AJA_CONTROL_H__

#ifdef WIN32

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <winsock2.h>
#include <ws2tcpip.h>

// Need to link with Ws2_32.lib, Mswsock.lib, and Advapi32.lib
#pragma comment (lib, "Ws2_32.lib")
#pragma comment (lib, "Mswsock.lib")
#pragma comment (lib, "AdvApi32.lib")

#pragma warning (disable:4996)
#define _CRT_NONSTDC_NO_WARNINGS 1
#define CLOSE_CONNECTION closesocket

#else

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#define SOCKET_ERROR -1
#define CLOSE_CONNECTION close

#endif

#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <vector>
#include <sstream>
#include <errno.h>
#include <fcntl.h>
#include <string.h>

namespace aja  { namespace examples { namespace restapi { namespace config_control {

//
// Example: Cross platform AJA control class using the AJA REST API via the /config handler
//

//
// Class: ConfigControl
//
// Description: A example class that provides a means to control an AJA REST Device. It's only dependency is on
//              basic socket APIs.
//

class ConfigControl
{
public:

	//
	// Method: AJAControl (and Destructor...)
	//
	// Description: Constructor for the AJAControl class
	//
	// Arguments:
	//
	//    host - the unit's IP address.
	//    port - The unit's port number (must be 80...)
	//    
	ConfigControl(std::string host, int port, std::string handler = "config") : _socket(-1), _host(host), _port(port), _handler(handler) {}
	virtual ~ConfigControl(void) { Disconnect(); }

	//
	// Method: SetParam
	//
	// Description: Changes the current value of a parameter.
	//
	// Arguments:
	//
	//    paramid - A valid parameter identifier name.
	//    value   - A string that contains the parameter's new value. NOTE: value must be urlencoded (see: http://en.wikipedia.org/wiki/Percent-encoding)
	//
	// Return Value: true if the parameter was set or false if an error occured (http status != 200)
	//    
	// Notes:
	//
	//    * All numeric values are base-10 signed strings, possibly fixed point.
	//    * Integer parameter values must be the numeric value.
	//    * Enum parameter values must be the enum's numeric value.
	//    * String parameter values are represented as-is.
	//
	bool SetParam(std::string paramid, std::string value)
	{
		if (!IsConnected() && !Connect())
			return false;

		std::string content = "?action=set&paramid=" + paramid + "&value=" + value;
		std::stringstream http_request;
		http_request
			<< "GET /" << _handler << content << " HTTP/1.1\r\n"
			<< "Host: " << _host << "\r\n"
			<< "Connection: keep-alive\r\n"
			<< "\r\n";

		int result = ::send(_socket, http_request.str().c_str(), http_request.str().length(), 0);
		if (result == SOCKET_ERROR)
		{
			perror("ConfigControl::ChangeParam - send failed");
			return false;
		}

		return (ReadHTTPResponse() == 200);
	}

	//
	// Method: GetParam
	//
	// Description: Obtains the current value of a parameter as a string.
	//
	// Arguments:
	//
	//    paramid - A valid parameter identifier name.
	//    value   - A reference to a string that will receive the parameter's value.
	//    alt     - A string specifying the desired mime-type used to represent the parameter's value.
	//    
	// Return Value: true if the parameter value was obtained or false if an error occured (http status != 200)
	//
	// Notes:
	//
	//    * All numeric values are base-10 signed strings, possibly fixed point.
	//    * Integer parameter values are represented simply as a numeric value, when alt=text-plain (default).
	//    * Enum parameter values are represented as the enum's numeric value, when alt=text-plain (default).
	//    * Supplying alt=json will return parameter information, including the value, represented as a json string.
	//
	bool GetParam(std::string paramid, std::string& value, std::string alt = "text-plain")
	{
		if (!IsConnected() && !Connect())
			return false;

		std::stringstream http_request;
		http_request 
			<< "GET /" << _handler << "?action=get&configid=0&paramid=" << paramid << "&alt=" << alt << " HTTP/1.1\r\n"
			<< "Host: " << _host << "\r\n"
			<< "Connection: keep-alive\r\n"
			<< "Content-type: application/x-www-form-urlencoded\r\n"
			<< "Content-length: 0\r\n"
			<< "\r\n";

		int result = ::send(_socket, http_request.str().c_str(), http_request.str().length(), 0);
		if (result == SOCKET_ERROR)
		{
			perror("ConfigControl::ChangeParam - send failed");
			return false;
		}

		std::vector<std::string> response_headers;
		std::string response_content;
		int http_status = ReadHTTPResponse(response_headers, response_content);
		if (http_status == 200)
		{
			value = response_content;
		}
		else
		{
			fprintf(stderr,"http_status = %d\n", http_status);
		}
		return (http_status == 200);
	}

	//
	// Method: IsConnected
	//
	// Description: Determines if the underlying socket is connected.
	//
	// Return Value: true if a valid socket connection exists
	//    
	bool IsConnected()
	{
		return (_socket != -1);
	}

	//
	// Method: Connect
	//
	// Description: Creates a socket and establishes a connection to the host/port for this KUMOControl instance
	//
	// Return Value: true if a connection was established, false otherwise.
	//
	// Notes:
	//
	//    * It is not strictly necessary to call this method since other class methods that require a connection
	//      will call it for you if no connection exists.
	//    
	bool Connect()
	{
		if ((_socket = ::socket(AF_INET, SOCK_STREAM, 0)) == -1)
		{
			perror("ConfigControl::Connect - socket");
			return false;
		}

		int flag = 1;
		if (::setsockopt(_socket, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(flag)) == -1)
		{
			perror("ConfigControl::Connect - setsockopt TCP_NODELAY");
			return false;
		}

		int on = 1;
		if (::setsockopt( _socket, SOL_SOCKET, SO_REUSEADDR, (char*)&on, sizeof(on)) == -1)
		{
			perror("ConfigControl::Connect - setsockopt SO_REUSEADDR");
			return false;
		}

#if defined(WIN32)
		struct addrinfo* result = NULL, hints;
		ZeroMemory( &hints, sizeof(hints) );
		hints.ai_family = AF_INET;
		hints.ai_socktype = SOCK_STREAM;
		hints.ai_protocol = IPPROTO_TCP;

		// Resolve the server address and port
		std::stringstream port_number;
		port_number << _port;
		int iResult = ::getaddrinfo(_host.c_str(), port_number.str().c_str(), &hints, &result);
		if ( iResult != 0 )
		{
			perror("ConfigControl::Connect - getaddrinfo failed");
			return false;
		}

		iResult = ::connect( _socket, result->ai_addr, (int)result->ai_addrlen);
        if (iResult == SOCKET_ERROR)
		{
			perror("ConfigControl::Connect - connect failed");
			return false;
        }
#else
		sockaddr_in addr;
		addr.sin_family = AF_INET;
		addr.sin_port = htons(_port);

		int result = ::inet_pton(AF_INET, _host.c_str(), &addr.sin_addr);
		if (result != 1)
		{
			perror("ConfigControl::Connect - inet_pton");
			 _socket = -1;
			return false;
		}

		if (::connect(_socket, (sockaddr*)&addr, sizeof(addr)) != 0)
		{
			perror("ConfigControl::Connect - connect");
			 _socket = -1;
			return false;
		}
#endif
		return true;
	}

	//
	// Method: Disconnect
	//
	// Description: Disconnects the socket connection for this KUMOControl instance
	//
	// Return Value: true
	//    
	bool Disconnect()
	{
		if (_socket == -1) return true;
		CLOSE_CONNECTION(_socket);
		_socket = -1;
		return true;
	}

protected:

	//
	// Method: ReadHTTPResponse
	//
	// Description: Helper function to read the HTTP response and obtain the HTTP status
	//
	// Return Value: The HTTP response status code.
	//    
	int ReadHTTPResponse()
	{
		std::vector<std::string> headers;
		std::string content;
		return ReadHTTPResponse(headers, content);
	}

	//
	// Method: ReadHTTPResponse
	//
	// Description: Helper function to read the HTTP response and obtain the HTTP status
	//
	// Arguments:
	//
	//    headers - A vector reference variable that will receive all HTTP headers sent by the server.
	//    content - A string reference variable that will receive the HTTP response content.
	//
	// Return Value: The HTTP response status code.
	//
	// Notes:
	//
	//    * Per the HTTP protocol, the server may request (via the HTTP response) that the current connection be closed.
	//      In this case, the connection will be closed and another one established as needed (see also: SetParam(), GetParam())
	//    
	int ReadHTTPResponse(std::vector<std::string>& headers, std::string& content)
	{
		// read the headers first
		// printf("headers:\n");
		int content_length = 0;
		bool chunked = false;
		bool connection_close = false;
		std::string header;
		int http_status = 0;
		do 
		{
			size_t pos;
			std::string header = Readline();
			headers.push_back(header);
			// printf("%s\n", header.c_str());

			// handle http status
			pos = header.find("HTTP/1.1 ");
			if (pos != std::string::npos)
			{
				if (sscanf(header.c_str(), "HTTP/1.1 %d", &http_status) != 1)
				{
					printf("bogus http status parse\n");
				}
			}

			// handle Connection: close
			pos = header.find("Connection: close");
			if (pos != std::string::npos)
			{
				connection_close = true;
			}

			// handle content-length
			pos = header.find("Content-Length: ");
			if (pos != std::string::npos)
			{
				if (sscanf(header.c_str(), "Content-Length: %d", &content_length) != 1)
				{
					printf("bogus content-length parse\n");
				}
			}

			// handle chunked transfer-encoding
			pos = header.find("Transfer-encoding: chunked");
			if (pos != std::string::npos)
			{
				chunked = true;
			}

			if (header == "")
				break;  // blank line (empty header) terminates headers...

		} while (true);

		if (content_length > 0)
		{
			// printf("content:\n");
			char buffer[MAX_RESPONSE_BUFFER_SIZE];
			memset(buffer, 0, MAX_RESPONSE_BUFFER_SIZE);
			::recv(_socket, buffer, content_length, 0);
			content = buffer;
			// printf(buffer);
			Readline(); // to clear last \r\n
			// printf("\n\n");
		}
		else if (chunked)
		{
			// printf("content:\n");
			std::string chunk_str;
			do
			{
				chunk_str = Readline();
				int chunk_size = 0;
				if (sscanf(chunk_str.c_str(), "%x", &chunk_size) == 1)
				{
					char buffer[MAX_RESPONSE_BUFFER_SIZE];
					memset(buffer, 0, MAX_RESPONSE_BUFFER_SIZE);
					::recv(_socket, buffer, chunk_size, 0);
					// printf(buffer);
					content.append(buffer);
				}
			} while (chunk_str != "0");
			Readline(); // to clear last \r\n
			// printf("\n\n");
		}

		if (connection_close)
		{
			Disconnect();
		}

		return http_status;
	}

	//
	// Method: Readline
	//
	// Description: Helper function to read a header from the HTTP response.
	//
	// Return Value: A string containing (presumably) an HTTP response header.
	//    
	std::string Readline()
	{
		std::string result;
		char ch;
		int n = 0;
		while ((n = ::recv(_socket, &ch, sizeof(ch), 0)) > 0)
		{
			if (ch == '\r') continue; // just ignore these...
			if (ch == '\n') break; // eol
			result += ch;
		}
		return result;
	}

	//
	// member data
	//

	int _socket;
	std::string _host;
	int _port;
	std::string _handler;
	static const int MAX_RESPONSE_BUFFER_SIZE = 4096;

#ifdef WIN32
	class WinSockHelper
	{
	public:
		WinSockHelper() {
			WSADATA wsaData;
			if (::WSAStartup(MAKEWORD(1,1), &wsaData) != 0) {
				fprintf(stderr, "WinSockHelper::WinSockHelper - WSAStartup failed.\n");
			}
		}

		virtual ~WinSockHelper() {
			::WSACleanup();
		}
	};
	WinSockHelper _winSockHelper;
#endif

};

}}}} // namespaces

#endif
