/*
	Copyright (C) 2012 AJA Video Systems, Inc.  Proprietary and Confidential information.  All rights reserved.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
	TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
	THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
	TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef __KUMO_CONTROL_H__
#define __KUMO_CONTROL_H__

#include "config_control.h"

namespace aja { namespace examples { namespace restapi { namespace kumo_control {

//
// Example: Cross platform KUMO control class using the KUMO REST API
//
// The kumo does it slightly differently - hence this derived class...
//

//
// Class: KUMOControl
//
// Description: A example class that provides a means to control a KUMO. It's only dependency is on
//              AJAControl
//

class KUMOControl : public aja::examples::restapi::config_control::ConfigControl
{
public:

	//
	// Method: KUMOControl (and Destructor...)
	//
	// Description: Constructor for the KUMOControl class
	//
	// Arguments:
	//
	//    host - the KUMO's IP address.
	//    port - The KUMO's port number (must be 80...)
	//    
	KUMOControl(std::string host, int port) : ConfigControl(host, port, "options") {}
	virtual ~KUMOControl(void) { Disconnect(); }

	//
	// Method: SetCrosspoint
	//
	// Description: Helper function to set the crosspoint between a source and a destination.
	//
	// Arguments:
	//
	//    destination - A destination number. The range depends on the KUMO model, 1-4 or 1-16.
	//    source      - A source number in the range 1-16.
	//
	// Return Value: true if crosspoint was changed, false otherwise
	//    
	bool SetCrosspoint(int destination, int source)
	{
		std::stringstream param_name;
		param_name << "eParamID_XPT_Destination" << destination << "_Status";
		std::stringstream value; value << source;
		return SetParam(param_name.str(), value.str());
	}

	//
	// Method: GetCrosspoint
	//
	// Description: Helper function to get the source assigned to a given destination.
	//
	// Arguments:
	//
	//    destination - A destination number. The range depends on the KUMO model, 1-4 or 1-16.
	//    source      - A reference to an integer variable that will receive the assigned source number (1-based source number. E.g. 1-16.)
	//
	// Return Value: true if crosspoint was obtained, false otherwise
	//    
	bool GetCrosspoint(int destination, int& source)
	{
		std::stringstream param_name;
		param_name << "eParamID_XPT_Destination" << destination << "_Status";
		std::string value;
		bool result = GetParam(param_name.str(), value);
		if (result)
		{
			source = atoi(value.c_str());
		}
		return result;
	}

	//
	// Method: SetParam
	//
	// Description: Changes the current value of a parameter.
	//
	// Arguments:
	//
	//    paramid - A valid parameter identifier name.
	//    value   - A string that contains the parameter's new value.
	//
	// Return Value: true if the parameter was set or false if an error occured (http status != 200)
	//    
	// Notes:
	//
	//    * All numeric values are base-10 signed strings, possibly fixed point.
	//    * Integer parameter values must be the numeric value.
	//    * Enum parameter values must be the enum's numeric value.
	//    * String parameter values are represented as-is.
	//
	bool SetParam(std::string paramid, std::string value)
	{
		if (!IsConnected() && !Connect())
			return false;

		std::string content = "paramName=" + paramid + "&newValue=" + value;
		std::stringstream http_request;
		http_request
			<< "POST /" << _handler << " HTTP/1.1\r\n"
			<< "Host: " << _host << "\r\n"
			<< "Connection: keep-alive\r\n"
			<< "Content-type: application/x-www-form-urlencoded\r\n"
			<< "Content-length: " << content.length() << "\r\n"
			<< "\r\n"
			<< content;

		int result = ::send(_socket, http_request.str().c_str(), http_request.str().length(), 0);
		if (result == SOCKET_ERROR)
		{
			perror("KUMOControl::ChangeParam - send failed");
			return false;
		}

		return (ReadHTTPResponse() == 200);
	}
};

}}}} // namespaces

#endif
